package edu.washington.cs.radishsim;

import static org.junit.Assert.assertEquals;

import org.junit.Before;
import org.junit.Test;

import edu.washington.cs.radishsim.Machine.SimulationMode;

public class UniproL3Tests {

   static Machine<RadishLine>            m;

   static final CpuId                    P0                   = new CpuId(0);
   static final int                      CORES                = 1;
   static final int                      L1_ASSOC             = 2;
   static final int                      LINE_SIZE            = 4;
   static final int                      L1_CACHE_SIZE        = 16;          // 4 blocks, 2 sets
   static final int                      L3_ASSOC             = 2;
   static final int                      L3_CACHE_SIZE        = 8;           // 2 blocks, 1 set
   static final int                      BITS_PER_LOCAL_EPOCH = 2;

   static {
      MemorySystemConstants.unsafeSetLineSize(LINE_SIZE);
      assert RadishSim.XASSERTS;
   }

   static CacheConfiguration<RadishLine> l1config             = new CacheConfiguration<RadishLine>() {
                                                                 {
                                                                    cacheSize = L1_CACHE_SIZE;
                                                                    lineSize = LINE_SIZE;
                                                                    assoc = L1_ASSOC;
                                                                    level = CacheLevel.L1;
                                                                 }
                                                              };
   static CacheConfiguration<RadishLine> l3config             = new CacheConfiguration<RadishLine>() {
                                                                 {
                                                                    cacheSize = L3_CACHE_SIZE;
                                                                    lineSize = LINE_SIZE;
                                                                    assoc = L3_ASSOC;
                                                                    level = CacheLevel.L3;
                                                                 }
                                                              };

   @Before
   public void setUp() {
      Machine.MachineParams<RadishLine> params = new DefaultTestingMachineParams() {

         @Override
         SimulationMode simulationMode() {
            return SimulationMode.BASELINE;
         }

         @Override
         int numProcessors() {
            return CORES;
         }

         @Override
         CacheConfiguration<RadishLine> l1config() {
            return l1config;
         }

         @Override
         int bitsPerLocalEpochOffset() {
            return BITS_PER_LOCAL_EPOCH;
         }

         @Override
         boolean useL3() {
            return true;
         }

         @Override
         CacheConfiguration<RadishLine> l3config() {
            return l3config;
         }
      };

      m = new Machine<RadishLine>(params);
   }

   // jargon:
   // Ad means data at location A
   // Amd=B means the metadata for data location A resides at location B

   @Test
   public void test_l3NoRadishRead() {
      Processor<RadishLine> p = m.getProc(P0);

      // bring 0d into the cache
      m.cacheRead(P0, 0, 1);
      assertEquals(1, p.stats.pc_l1d.pc_ReadMisses.get());

      // fetch 16d and 32d which map to set 0; this evicts 0d to the L3
      for ( int i = 1; i <= L1_ASSOC; i++ ) {
         m.cacheRead(P0, L1_CACHE_SIZE * i, 1);
         assertEquals(1 + i, p.stats.pc_l1d.pc_ReadMisses.get());
      }

      // fetch 0d again, from the L3
      m.cacheRead(P0, 0, 1);
      assertEquals(1, p.stats.pc_l3d.pc_ReadHits.get());
   }

   @Test
   public void test_l3NoRadishRead1() {
      Processor<RadishLine> p = m.getProc(P0);

      // bring 0d into the cache
      m.cacheRead(P0, 0, 1);
      assertEquals(1, p.stats.pc_l1d.pc_ReadMisses.get());

      // fetch things which map to set 0; this evicts 0d to memory
      for ( int i = 1; i <= L1_ASSOC + L3_ASSOC; i++ ) {
         m.cacheRead(P0, L1_CACHE_SIZE * i, 1);
         assertEquals(1 + i, p.stats.pc_l1d.pc_ReadMisses.get());
      }

      // fetch 0d again; it's a miss
      m.cacheRead(P0, 0, 1);
      assertEquals(2 + L1_ASSOC + L3_ASSOC, p.stats.pc_l1d.pc_ReadMisses.get());
   }

   @Test
   public void test_l3NoRadishWrite() {
      Processor<RadishLine> p = m.getProc(P0);

      // bring 0d into the cache
      m.cacheWrite(P0, 0, 1);
      assertEquals(1, p.stats.pc_l1d.pc_WriteMisses.get());

      // fetch 16d and 32d which map to set 0; this evicts 0d to the L3
      m.cacheWrite(P0, L1_CACHE_SIZE, 1);
      m.cacheWrite(P0, L1_CACHE_SIZE * 2, 1);
      assertEquals(3, p.stats.pc_l1d.pc_WriteMisses.get());

      // fetch 0d again, from the L3
      m.cacheRead(P0, 0, 1);
      assertEquals(3, p.stats.pc_l1d.pc_WriteMisses.get());
      assertEquals(1, p.stats.pc_l3d.pc_ReadHits.get());
   }

   @Test
   public void test_l3TagBug() {
      Processor<RadishLine> p = m.getProc(P0);

      // bring 4d into the cache
      m.cacheWrite(P0, 4, 1);
      assertEquals(1, p.stats.pc_l1d.pc_WriteMisses.get());

      // fetch 20d and 36d which map to set 1; this evicts 4d to the L3
      m.cacheWrite(P0, 4 + L1_CACHE_SIZE, 1);
      m.cacheWrite(P0, 4 + (L1_CACHE_SIZE * 2), 1);
      assertEquals(3, p.stats.pc_l1d.pc_WriteMisses.get());

      // fetch 4d again, from the L3
      m.cacheRead(P0, 4, 1);
      assertEquals(3, p.stats.pc_l1d.pc_WriteMisses.get());
      assertEquals(1, p.stats.pc_l3d.pc_ReadHits.get());
   }

   @Test
   public void test_l3SimpleRadish() {
      Machine.MachineParams<RadishLine> params = new DefaultTestingMachineParams() {
         @Override
         SimulationMode simulationMode() {
            return SimulationMode.RADISH;
         }

         @Override
         int numProcessors() {
            return CORES;
         }

         @Override
         CacheConfiguration<RadishLine> l1config() {
            return l1config;
         }

         @Override
         int bitsPerLocalEpochOffset() {
            return BITS_PER_LOCAL_EPOCH;
         }

         @Override
         boolean useL3() {
            return true;
         }

         @Override
         CacheConfiguration<RadishLine> l3config() {
            return l3config;
         }
      };

      Machine<RadishLine> machine = new Machine<RadishLine>(params);

      Processor<RadishLine> p = machine.getProc(P0);

      // bring 0d,0md=0 into same set
      machine.cacheRead(P0, 0, 1);
      assertEquals(1, p.stats.pc_l1md.pc_ReadMisses.get());
      assertEquals(0, p.stats.pc_MetadataEvictions.get());

      // fetch 16d,16md=32, which map to set 0 and evict 0d,0md=0 to L3
      machine.cacheRead(P0, L1_CACHE_SIZE, 1);
      assertEquals(2, p.stats.pc_l1md.pc_ReadMisses.get());
      assertEquals(0, p.stats.pc_MetadataEvictions.get());

      // fetch 0d,0md=0 from L3, evicting 16d,16md=32 to L3
      machine.cacheRead(P0, 0, 1);
      assertEquals(1, p.stats.pc_l3md.pc_ReadHits.get());
      assertEquals(0, p.stats.pc_MetadataEvictions.get());
   }

}
