#ifndef EVENT_HPP_
#define EVENT_HPP_

#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <map>
#include <set>
#include <string>

using namespace std;

#include <assert.h>
#include <stdint.h>

inline void endian_swap(int16_t& x) {
    x = (x>>8) |
        (x<<8);
}

inline void endian_swap(int32_t& x) {
    x = (x>>24) |
        ((x<<8) & 0x00FF0000) |
        ((x>>8) & 0x0000FF00) |
        (x<<24);
}

inline void endian_swap(uint64_t& x) {
    x = (x>>56) |
        ((x<<40) & 0x00FF000000000000) |
        ((x<<24) & 0x0000FF0000000000) |
        ((x<<8)  & 0x000000FF00000000) |
        ((x>>8)  & 0x00000000FF000000) |
        ((x>>24) & 0x0000000000FF0000) |
        ((x>>40) & 0x000000000000FF00) |
        (x<<56);
}

#define INVALID_FRONTEND_THREADID (-1)
typedef int16_t frontend_tid_t;

enum EventType {
  INVALID_EVENT = 0,
  ROI_START = 1, ROI_FINISH = 2, /** Tracking the Parsec region-of-interest */
  THREAD_START = 3, /** When a new thread starts running, so we can setup its resources */
  THREAD_FINISH = 4, /** When a thread exits, so we can teardown its resources */
  THREAD_BLOCKED = 5, /** When a thread is about to block inside the kernel, e.g. when doing a join */
  THREAD_UNBLOCKED = 6, /** When a thread returns from blocking inside the kernel, e.g. after having joined */
  MEMORY_READ = 7, MEMORY_WRITE = 8,
  MEMORY_ALLOCATION = 9, MEMORY_FREE = 10,
  BASIC_BLOCK = 11, /** Insn counting */
  HAPPENS_BEFORE_SOURCE = 12, HAPPENS_BEFORE_SINK = 13, /** All sync ops are mapped down to HB edges */
  COMMUNICATION_TEST = 14 /** Used to verify that frontend can talk to simulator */
};

class Event {
#ifdef SIMULATOR_FRONTEND
private:
  /** Which thread last accessed each sync object. Used to provide source
   * information on HB-sink events. */
  static map<ADDRINT, frontend_tid_t> s_WhoLastAccessed;
  /** Lock for s_WhoLastAccessed */
  static PinLock sl_LastAccessed;
#endif

public:
  EventType m_type;
  frontend_tid_t m_tid;

  // MEMORY_READ, MEMORY_WRITE, MEMORY_ALLOCATION, MEMORY_FREE
  uint64_t m_addr;
  int32_t m_memOpSize;
  bool m_stackRef;

  // HAPPENS_BEFORE_SOURCE, HAPPENS_BEFORE_SINK
  bool m_isLifeLock;
  uint64_t m_syncObject;
  uint64_t m_logicalTime; /** filled in by the simulator */

  /** The thread that was the source of this HB-sink event. INVALID_FRONTEND_THREADID if
  there is no such thread, e.g. the very first time this lock is acquired. */
  frontend_tid_t m_hbSourceThread; // HAPPENS_BEFORE_SINK

  int8_t m_insnCount; // BASIC_BLOCK

#ifdef SIMULATOR_FRONTEND
  /** Serialize an event on the given output stream. This *must be kept in sync*
   * with the deserialization code in the simulator. */
  void send(ofstream& os) {
    char t = m_type;
    os.write( &t, 1 );
    endian_swap( m_tid );
    os.write( (char*) &m_tid, 2 );
    endian_swap( m_addr );
    os.write( (char*) &m_addr, 8 );
    endian_swap( m_memOpSize );
    os.write( (char*) &m_memOpSize, 4 );
    char b = 0; // pack into a bitfield
    if (m_stackRef) b |= 1;
    if (m_isLifeLock) b |= 2;
    os.write( &b, 1 );
    endian_swap( m_syncObject );
    os.write( (char*) &m_syncObject, 8 );
    // m_logicalTime lives only on the simulator side
    endian_swap( m_hbSourceThread );
    os.write( (char*) &m_hbSourceThread, 2 );
    os.write( (char*) &m_insnCount, 1 );
  }

  static Event MemoryEvent( frontend_tid_t tid, EventType typ, uint64_t addr,
                            int32_t memOpSize, bool stackRef ) {
    assert( MEMORY_READ == typ || MEMORY_WRITE == typ );
    Event e = Event( tid, typ );
    e.m_addr = addr;
    e.m_memOpSize = memOpSize;
    e.m_stackRef = stackRef;
    return e;
  }

  static Event AllocationEvent( frontend_tid_t tid, EventType typ, uint64_t startAddr, int32_t extent=0) {
    assert( MEMORY_ALLOCATION == typ || MEMORY_FREE == typ );
    if ( 0 == extent ) {
      assert( MEMORY_FREE == typ );
    }
    Event e = Event( tid, typ );
    e.m_addr = startAddr;
    e.m_memOpSize = extent;
    return e;
  }

  static Event SyncSourceEvent( frontend_tid_t tid, EventType typ, uint64_t syncObject, bool lifelock ) {
    assert( HAPPENS_BEFORE_SOURCE == typ );
    Event e = Event( tid, typ );
    e.m_syncObject = syncObject;
    e.m_isLifeLock = lifelock;

    sl_LastAccessed.lock();
    s_WhoLastAccessed[syncObject] = tid;
    sl_LastAccessed.unlock();

    return e;
  }

  static Event SyncSinkEvent( frontend_tid_t tid, EventType typ, uint64_t syncObject, bool lifelock ) {
    assert( HAPPENS_BEFORE_SINK == typ );
    Event e = Event( tid, typ );
    e.m_syncObject = syncObject;
    e.m_isLifeLock = lifelock;

    THREADID lastAccessor = INVALID_FRONTEND_THREADID;
    sl_LastAccessed.lock();
    if ( s_WhoLastAccessed.count(syncObject) > 0 ) {
      lastAccessor = s_WhoLastAccessed.at( syncObject );
    }
    sl_LastAccessed.unlock();

    e.m_hbSourceThread = lastAccessor;
    return e;
  }

  static Event BasicBlockEvent( frontend_tid_t tid, EventType typ, int8_t insnCount ) {
    assert( BASIC_BLOCK == typ );
    Event e = Event( tid, typ );
    e.m_insnCount = insnCount;
    return e;
  }

  static Event ThreadEvent( frontend_tid_t tid, EventType typ ) {
    assert( ROI_START == typ || ROI_FINISH == typ || THREAD_START == typ ||
            THREAD_FINISH == typ || THREAD_BLOCKED == typ || THREAD_UNBLOCKED == typ );
    return Event( tid, typ );
  }

  static Event CommTestEvent() {
    Event e = Event( 10, COMMUNICATION_TEST );
    e.m_addr = 11;
    e.m_memOpSize = 12;
    e.m_stackRef = true; e.m_isLifeLock = true; // encoded as 3 in the bitfield
    e.m_syncObject = 14;
    e.m_hbSourceThread = 15;
    e.m_insnCount = 16;
    return e;
  }
#endif

  /** This no-arg ctor is needed only for the lockfree ringbuffer. */
  Event() {
    clear();
  }

  /** reset this Event */
  void clear() {
    m_type = INVALID_EVENT;
    m_tid = -1;
    m_addr = 0;
    m_memOpSize = 0;
    m_stackRef = false;
    m_syncObject = 0;
    m_hbSourceThread = -1;
    m_insnCount = 0;
    m_logicalTime = 0;
    m_isLifeLock = false;
  }

  string toString() {
    string name;
    stringstream ss;

    switch ( m_type ) {
    case INVALID_EVENT:
      ss << "INVALID event";
      ss << " tid=" << m_tid;
      ss << " addr=" << m_addr;
      ss << " size=" << m_memOpSize;
      ss << " stack=" << m_stackRef;
      ss << " syncObj=" << m_syncObject;
      ss << " sourceTid=" << m_hbSourceThread;
      ss << " icount=" << m_insnCount;
      break;

    case ROI_START:
      name = "ROI_start";
      goto PrintThreadEvent;
    case ROI_FINISH:
      name = "ROI_finish";
      goto PrintThreadEvent;
    case THREAD_BLOCKED:
      name = "thread_blocked";
      goto PrintThreadEvent;
    case THREAD_UNBLOCKED:
      name = "thread_unblocked";
      goto PrintThreadEvent;
    case THREAD_START:
      name = "thread_start";
      goto PrintThreadEvent;
    case THREAD_FINISH:
      name = "thread_finish";
      PrintThreadEvent: ss << name << ", tid=" << m_tid;
      break;

    case MEMORY_READ:
      name = "read";
      goto PrintMemEvent;
    case MEMORY_WRITE:
      name = "write";
      PrintMemEvent: ss << name << ", tid=" << m_tid << ", size=" << m_memOpSize
          << ", stack=" << m_stackRef;
      break;

    case MEMORY_ALLOCATION:
      name = "alloc";
      goto PrintAllocEvent;
    case MEMORY_FREE:
      name = "free";
      PrintAllocEvent: ss << name << ", tid=" << m_tid << ", addr=0x" << hex << m_addr << dec << ", size=" << m_memOpSize ;
      break;

    case BASIC_BLOCK:
      ss << "basicblock" << ", tid=" << m_tid << ", insnCount=" << m_insnCount;
      break;

    case HAPPENS_BEFORE_SOURCE:
      name = "HB_source";
      goto PrintSyncEvent;
    case HAPPENS_BEFORE_SINK:
      name = "HB_sink";
      PrintSyncEvent: ss << name << ", tid=" << m_tid << ", syncObject=0x" << hex << m_syncObject << dec;
      if ( m_hbSourceThread != INVALID_FRONTEND_THREADID ) {
        ss << ", hbSourceTid=" << m_hbSourceThread;
      }
      break;
    case COMMUNICATION_TEST:
      name = "comm test";
      break;

    default:
      assert(false);
    }

    return ss.str();
  }

private:
  /** Private ctor */
  Event( frontend_tid_t tid, EventType typ ) {
    clear();
    m_tid = tid;
    m_type = typ;
  }

};

#endif /* EVENT_HPP_ */
